from growatt_client import GrowattClient


def gen_markup():
    client = GrowattClient()

    print("| Attribute | Register | Unit | Calculated | Description | Misc |")
    print("| --- | ---: | --- | --- | --- | --- |")

    attributes = client.get_attributes()
    attributes.sort(key=lambda x: x["name"])
    for attr in attributes:
        name = attr["name"]
        d = attr["description"]
        u = attr["unit"]
        calculated = "(calculated)" if "template" in attr else ""
        if calculated:
            m = attr["template"]
            print(f"| {name} | - | {u} | &check; | {d} | {m} |")
        else:
            r = attr["pos"]
            print(f"| {name} | {r} |{u} | | {d} | |")


if __name__ == "__main__":
    gen_markup()



"""
Python wrapper for getting data asynchronously from Growatt inverters
via serial usb RS232/RS485 connection and modbus RTU protocol.
"""
import logging
import os

from pymodbus.client.serial import AsyncModbusSerialClient as ModbusClient

from .const import (
    ATTRIBUTES,
    ATTRIBUTES_GROUPED,
    ATTRIBUTE_TEMPALTES,
    DEFAULT_ADDRESS,
    DEFAULT_PORT,
    DOUBLE_BYTE,
    INT_BYTE,
    SINGLE_BYTE,
)


def get_from_single_byte(rr, index, scale=0.1):
    """Read and scale single to float."""
    return round(float(rr.registers[index]) * scale, 1)


def get_from_double_byte(rr, index, scale=0.1):
    """Read and scale double to float."""
    return round(
        float((rr.registers[index] << 16) + rr.registers[index + 1]) * scale, 1
    )


def get_byte(rr, index):
    return rr.registers[index]


def get_value(res, index, type, scale):
    if type == INT_BYTE:
        return get_byte(res, index)
    if type == SINGLE_BYTE:
        return get_from_single_byte(res, index, scale)
    if type == DOUBLE_BYTE:
        return get_from_double_byte(res, index, scale)
    return 0


def get_string(res, index, length):
    string = ""
    for pos in range(0, length):
        string += str(
            chr(res.registers[index + pos] >> 8)
            + chr(res.registers[index + pos] & 0x000000FF)
        )
    return string


class GrowattClient:
    """Main class to communicate with the Growatt inverter."""

    def __init__(
        self, port=DEFAULT_PORT, address=DEFAULT_ADDRESS, logger=None
    ):
        """Initialize."""
        if logger is None:
            self._logger = logging.getLogger(__name__)
            self._logger.addHandler(logging.NullHandler())
        else:
            self._logger = logger

        # usb port
        self._port = port
        self._address = address

        if not os.path.exists(self._port):
            self._logger.debug(f"USB port {self._port} is not available")
            raise PortException(f"USB port {self._port} is not available")

        # Modbus serial rtu communication client
        self._client = ModbusClient(
            port=port,
            baudrate=9600,
            stopbits=1,
            parity="N",
            bytesize=8,
            timeout=1,
        )

        self._serial_number = ""
        self._model_number = ""
        self._firmware = ""

        self._logger.debug(
            f"GrowattClient initialized with usb port {self._port}"
        )

    async def async_update(self):
        """
        Read Growatt data.

        Modbus rtu information from
        "Growatt Inverter Modbus RTU Protocol V1.20 2020-04-28".
        The availability of the attributes depends
        on the firmware version of your inverter.
        """

        data = {}
        await self.update_hardware_info()

        for group in ATTRIBUTES_GROUPED:
            pos = group["pos"]
            values = group["values"]
            self._logger.debug(group)

            if not await self._client.connect():
                self._logger.debug("Modbus connection failed.")
                raise ModbusException("Modbus connection failed.")

            registers = await self._client.read_input_registers(
                pos, group["length"], slave=self._address
            )

            if registers.isError():
                await self._client.close()
                self._logger.debug(f"Modbus read failed for registers {pos}.")
                raise ModbusException(
                    f"Modbus read failed for registers {pos}."
                )
            for value in values:
                val = get_value(
                    registers,
                    value["pos"] - pos,
                    value["type"],
                    value["scale"],
                )
                data[value["name"]] = val

            await self._client.close()

        for value in ATTRIBUTE_TEMPALTES:
            templ = value["template"].format_map(data)
            val = round(eval(templ), 1)
            self._logger.debug(
                "Converting template: %s \n -> to: %s = %f",
                value["template"],
                templ,
                val,
            )
            data[value["name"]] = val

        return data

    async def update_hardware_info(self):
        if self._serial_number == "":
            if not await self._client.connect():
                self._logger.debug("Modbus connection failed.")
                raise ModbusException("Modbus connection failed.")

            # Assuming the serial number doesn't change, it is read only once
            registers = await self._client.read_holding_registers(
                0, 30, slave=self._address
            )
            if registers.isError():
                await self._client.close()
                self._logger.debug("Modbus read failed for holding registers.")
                raise ModbusException(
                    "Modbus read failed for holding registers."
                )

            self._firmware = get_string(registers, 9, 3)
            self._serial_number = get_string(registers, 23, 5)

            mo = (registers.registers[28] << 16) + registers.registers[29]
            self._model_number = (
                "T"
                + str((mo & 0xF00000) >> 20)
                + " Q"
                + str((mo & 0x0F0000) >> 16)
                + " P"
                + str((mo & 0x00F000) >> 12)
                + " U"
                + str((mo & 0x000F00) >> 8)
                + " M"
                + str((mo & 0x0000F0) >> 4)
                + " S"
                + str((mo & 0x00000F))
            )

            self._logger.debug(
                (
                    f"Growatt serial number {self._serial_number} "
                    f"is model {self._model_number} "
                    f"and has firmware {self._firmware}."
                )
            )

            # Read time
            # self._logger.info(
            #     (
            #         f"{registers.registers[45]}-"
            #         f"{registers.registers[46]:02d}-"
            #         f"{registers.registers[47]:02d} "
            #         f"{registers.registers[48]:02d}:"
            #         f"{registers.registers[49]:02d}:"
            #         f"{registers.registers[50]:02d} "
            #         # f"day: {registers.registers[51]} "
            #     )
            # )

            await self._client.close()

    def get_attributes(self):
        return ATTRIBUTES + ATTRIBUTE_TEMPALTES

    def get_attribute(self, name):
        for a in self.get_attributes():
            if a["name"] == name:
                return a

    def get_serial_number(self):
        return self._serial_number

    def get_firmware(self):
        return self._firmware

    def get_model_number(self):
        return self._model_number


class PortException(Exception):
    """Raised when the USB port in not available."""

    def __init__(self, status):
        """Initialize."""
        super(PortException, self).__init__(status)
        self.status = status


class ModbusException(Exception):
    """Raised when the Modbus communication has error."""

    def __init__(self, status):
        """Initialize."""
        super(ModbusException, self).__init__(status)
        self.status = status



import asyncio
import logging
from sys import argv

from growatt_client import GrowattClient

# defaults
# USB port of RS232/RS485 converter
DEFAULT_PORT = "/dev/ttyUSB0"
# Growatt modbus address
DEFAULT_ADDRESS = 0x01

logging.basicConfig(level=logging.INFO)


async def run_async_client():
    port = str(argv[1]) if len(argv) > 1 else DEFAULT_PORT
    address = int(argv[2]) if len(argv) > 2 else DEFAULT_ADDRESS
    client = GrowattClient(port, address)

    data = await client.async_update()
    ser = client.get_serial_number()

    logging.info(
        f" Serial number: {ser} "
        f"Firmware: {client.get_firmware()} "
        f"Model Number: {client.get_model_number()}"
    )

    logging.debug(f"Sensors data: {data}")
    for key in sorted(data):
        desc = client.get_attribute(key)
        value = data[key]
        d = desc["description"]
        u = desc["unit"]
        calculated = "(calculated)" if "template" in desc else ""
        logging.info(f"{d} {value} {u} {calculated}")

    # except Exception as error:
    #     logging.error("Error: " + repr(error))


if __name__ == "__main__":
    asyncio.run(run_async_client())



"""Constants for GrowattClient library."""

# Defaults
DEFAULT_PORT = "/dev/ttyUSB0"
DEFAULT_ADDRESS = 0x1

# Word type
INT_BYTE = "int_byte"
SINGLE_BYTE = "single_byte"
DOUBLE_BYTE = "double_byte"

# Unit of measurement
ELECTRICAL_POTENTIAL_VOLT = "V"
ELECTRICAL_CURRENT_AMPERE = "A"
POWER_KILO_WATT = "kW"
REACTIVE_POWER_VAR = "var"
TIME_HOURS = "h"
ENERGY_KILO_WATT_HOUR = "kWh"
REACTIVE_ENERGY_KILO_VAR_HOUR = "kvarh"
FREQUENCY_HERTZ = "Hz"
TEMP_CELSIUS = "°C"


def create_value(name, pos, unit, desc, type=DOUBLE_BYTE, scale=0.1):
    return {
        "name": name,
        "pos": pos,
        "type": type,
        "unit": unit,
        "description": desc,
        "scale": scale,
    }


def create_template(name, unit, desc, template):
    """Creates a nice template value"""
    template = " ".join(template.split())
    value = {
        "name": name,
        "unit": unit,
        "description": desc,
        "template": template,
    }
    return value


def group_values(values):
    values.sort(key=lambda x: x["pos"])
    groups = []
    pos = -200

    for value in values:
        if pos + 100 < value["pos"] or (pos < 0 and value["pos"] >= 0):
            pos = value["pos"]
            group_list = []
            group = {}
            group["pos"] = pos
            group["values"] = group_list
            groups.append(group)
        group_list.append(value)
        group["length"] = (
            value["pos"] - pos + (2 if value["type"] == DOUBLE_BYTE else 1)
        )
    return groups


ATTRIBUTES = [
    create_value("pv1", 5, POWER_KILO_WATT, "PV1", DOUBLE_BYTE, 0.0001),
    create_value(
        "pv1_voltage",
        3,
        ELECTRICAL_POTENTIAL_VOLT,
        "PV1 voltage",
        SINGLE_BYTE,
    ),
    create_value(
        "pv1_today",
        59,
        ENERGY_KILO_WATT_HOUR,
        "PV1 today",
    ),
    create_value(
        "pv1_lifetime",
        61,
        ENERGY_KILO_WATT_HOUR,
        "PV1 total",
    ),
    create_value("pv2", 9, POWER_KILO_WATT, "PV2", DOUBLE_BYTE, 0.0001),
    create_value(
        "pv2_voltage",
        7,
        ELECTRICAL_POTENTIAL_VOLT,
        "PV2 voltage",
        SINGLE_BYTE,
    ),
    create_value(
        "pv2_today",
        63,
        ENERGY_KILO_WATT_HOUR,
        "PV2 today",
    ),
    create_value(
        "pv2_lifetime",
        65,
        ENERGY_KILO_WATT_HOUR,
        "PV2 total",
    ),
    create_value(
        "solar_generation",
        1,
        POWER_KILO_WATT,
        "All PV generation",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "solar_generation_lifetime",
        91,
        ENERGY_KILO_WATT_HOUR,
        "All PV generation total",
    ),
    # Battery
    create_value(
        "statement_of_charge",
        1014,
        "%",
        "Statement of charge",
        INT_BYTE,
        1,
    ),
    create_value(
        "battery_voltage",
        1013,
        ELECTRICAL_POTENTIAL_VOLT,
        "Battery voltage",
        SINGLE_BYTE,
    ),
    create_value(
        "battery_charge",
        1011,
        POWER_KILO_WATT,
        "Battery charging",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "battery_charge_today",
        1056,
        ENERGY_KILO_WATT_HOUR,
        "Battery charged today",
    ),
    create_value(
        "battery_charge_lifetime",
        1058,
        ENERGY_KILO_WATT_HOUR,
        "Battery charged total",
    ),
    create_value(
        "battery_discharge",
        1009,
        POWER_KILO_WATT,
        "Battery discharging",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "battery_discharge_today",
        1052,
        ENERGY_KILO_WATT_HOUR,
        "Battery discharged today",
    ),
    create_value(
        "battery_discharge_lifetime",
        1054,
        ENERGY_KILO_WATT_HOUR,
        "Battery discharged total",
    ),
    # Load consumtion
    create_value(
        "local_load",
        1037,
        POWER_KILO_WATT,
        "Inverter local load",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "local_load_today",
        1060,
        ENERGY_KILO_WATT_HOUR,
        "Inverter local load today",
    ),
    create_value(
        "local_load_lifetime",
        1062,
        ENERGY_KILO_WATT_HOUR,
        "Inverter local load total",
    ),
    # Export to grid
    create_value(
        "export_to_grid",
        1029,
        POWER_KILO_WATT,
        "Export to grid",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "export_to_grid_today",
        1048,
        ENERGY_KILO_WATT_HOUR,
        "Export to grid today",
    ),
    create_value(
        "export_to_grid_lifetime",
        1050,
        ENERGY_KILO_WATT_HOUR,
        "Export to grid total",
    ),
    # Import from grid
    create_value(
        "import_from_grid",
        1021,
        POWER_KILO_WATT,
        "Import from grid",
        DOUBLE_BYTE,
        0.0001,
    ),
    create_value(
        "import_from_grid_today",
        1044,
        ENERGY_KILO_WATT_HOUR,
        "Import from grid today",
    ),
    create_value(
        "import_from_grid_lifetime",
        1046,
        ENERGY_KILO_WATT_HOUR,
        "Import from grid total",
    ),
    create_value(
        "system_production_today",
        1137,
        ENERGY_KILO_WATT_HOUR,
        "System production today",
    ),
    create_value(
        "system_production_lifetime",
        1139,
        ENERGY_KILO_WATT_HOUR,
        "System production total",
    ),
    create_value(
        "grid_voltage",
        38,
        ELECTRICAL_POTENTIAL_VOLT,
        "Grid voltage",
        SINGLE_BYTE,
    ),
    # create_value(
    #     "extra_inverter_power_today",
    #     1133,
    #     ENERGY_KILO_WATT_HOUR,
    #     "Extra inverter Power today",
    # ),
    # create_value(
    #     "extra_inverter_power_lifetime",
    #     1135,
    #     ENERGY_KILO_WATT_HOUR,
    #     "Extra inverter Power total",
    # ),
    create_value(
        "grid_frequency",
        37,
        FREQUENCY_HERTZ,
        "Grid frequency",
        SINGLE_BYTE,
        0.01
    ),
    create_value(
        "inverter_temperature1",
        93,
        TEMP_CELSIUS,
        "Inverter temperature",
        SINGLE_BYTE,
    ),
    create_value(
        "inverter_temperature2",
        94,
        TEMP_CELSIUS,
        "The inside IPM in inverter Temperature",
        SINGLE_BYTE,
    ),
    create_value(
        "inverter_temperature3",
        1040,
        TEMP_CELSIUS,
        "Battery temperature",
        SINGLE_BYTE,
    ),
]

ATTRIBUTES_GROUPED = group_values(ATTRIBUTES)


ATTRIBUTE_TEMPALTES = [
    create_template(
        "solar_generation_today",
        ENERGY_KILO_WATT_HOUR,
        "All PV generation today",
        "{pv2_today} + {pv1_today}",
    ),
    create_template(
        "consumption",
        POWER_KILO_WATT,
        "Consumption",
        "{solar_generation} + {battery_discharge} + {import_from_grid} - \
            {export_to_grid} - {battery_charge}",
    ),
    create_template(
        "consumption_today",
        ENERGY_KILO_WATT_HOUR,
        "Consumption today",
        "{solar_generation_today} + {battery_discharge_today} + \
            {import_from_grid_today} - {export_to_grid_today} - \
            {battery_charge_today}",
    ),
    create_template(
        "consumption_lifetime",
        ENERGY_KILO_WATT_HOUR,
        "Consumption total",
        "{solar_generation_lifetime} + {battery_discharge_lifetime} + \
            {import_from_grid_lifetime} - {export_to_grid_lifetime} - \
            {battery_charge_lifetime}",
    ),
]



import time
import json
from configparser import ConfigParser
import paho.mqtt.client as mqtt
import logging
import asyncio

from sys import argv

from growatt_client import GrowattClient

# USB port of RS232/RS485 converter
DEFAULT_PORT = "/dev/ttyUSB0"
# Growatt modbus address
DEFAULT_ADDRESS = 0x00

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger("Mqtt-example")

connected = False


def on_connect(client, userdata, flags, rc):
    if rc == 0:
        logger.info(f"connected OK to MQTT Returned code={rc}")
        global connected
        connected = True
    else:
        logger.warning(f"Bad connection Returned code={rc}")


port = str(argv[1]) if len(argv) > 1 else DEFAULT_PORT
address = int(argv[2]) if len(argv) > 2 else DEFAULT_ADDRESS
client = GrowattClient(port, address)

mqtt_client = mqtt.Client("inverter")
mqtt_client.on_connect = on_connect
mqtt_client.connect("aceone.se")
mqtt_client.loop_start()

while not connected:
    time.sleep(2)

payload = {"name": "Growatt MQTT", "unique_id": "growatt_inverter"}
logger.info(payload)
mqtt_client.publish(
    "homeassistant/sensor/growattmqtt/config",
    payload=json.dumps(payload),
    retain=True,
)

# NAMES = [
#     "pv1_wattage",
#     "pv1_energy_today",
#     "pv1_energy_lifetime",
#     "pv2_wattage",
#     "pv2_energy_today",
#     "pv2_energy_lifetime",
#     ]
# for name in NAMES:
#     mqtt_client.publish(
#         "homeassistant/sensor/growatt_" + name + "/config",
#         payload="",
#         # payload=json.dumps(payload),
#         retain=True,
#     )



for reg in client.get_attributes():
    name = reg["name"]
    unit = reg["unit"]
    desc = reg["description"]
    if unit == "kWh":
        class_type = "energy"
    elif unit == "kW":
        class_type = "power"
    elif unit == "v":
        class_type = "battery"
    elif unit == "v":
        class_type = "voltage"
    elif unit == "ts":
        class_type = "timestamp"

    unique_id = "growatt_inverter_" + name
    topic = "inverter/growattmqtt/" + name

    payload = {
        "device_class": class_type,
        "name": desc,
        "state_topic": topic,
        "unit_of_measurement": unit,
        "unique_id": unique_id,
        "device": {
            "identifiers": ["growatt_inverter"],
            "name": "growatt_inverter",
        },
    }

    logger.info(payload)

    mqtt_client.publish(
        "homeassistant/sensor/growatt_" + name + "/config",
        payload=json.dumps(payload),
        retain=True,
    )


async def main():
    try:
        while True:
            regs = await client.async_update()
            for key, value in regs.items():
                logger.info(f"name: {key}, value: {value}")
                mqtt_client.publish(
                    "inverter/growattmqtt/" + key, payload=value
                )
            time.sleep(30)
            logger.info("")
    except Exception as error:
        logger.error("Error: " + repr(error))


loop = asyncio.get_event_loop()
loop.run_until_complete(main())
loop.close()
